/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2014 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        28. August 2014
 * $Revision:    V2.00
 *
 * Driver:       Driver_USBD0
 * Configured:   via RTE_Device.h configuration file 
 * Project:      USB Device Driver for SiliconLabs EFM32GGxxx
 * -----------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 2.00
 *    Updated to 2.00 API
 *  Version 1.04
 *    Multiple packet read
 *  Version 1.03
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

/* NOTE:
 *  - Isochronous endpoints are not supported
 */

#include <stdint.h>
#include <string.h>
#include "cmsis_os.h"

#include "Driver_USBD.h"

#include "em_device.h"
#include "em_cmu.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

/* USBD Driver ****************************************************************/

#define ARM_USBD_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(2,0) /* USBD driver version */

// Driver Version
static const ARM_DRIVER_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

// Driver Capabilities
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  0,  // VBUS detection
  0,  // Signal VBUS On event
  0,  // Signal VBUS Off event
};

// Number of endpoints
#if   (RTE_USB_DEV_EP & 0x40)
#define USBD_EP_NUM      6
#elif (RTE_USB_DEV_EP & 0x20)
#define USBD_EP_NUM      5
#elif (RTE_USB_DEV_EP & 0x10)
#define USBD_EP_NUM      4
#elif (RTE_USB_DEV_EP & 0x08)
#define USBD_EP_NUM      3
#elif (RTE_USB_DEV_EP & 0x04)
#define USBD_EP_NUM      2
#elif (RTE_USB_DEV_EP & 0x02)
#define USBD_EP_NUM      1
#else
#define USBD_EP_NUM      0
#endif

// Interrupt IN Endpoint intermediate buffers
#define USB_INT_IN_EP        ((RTE_USB_DEV_EP_INT >> 16) & RTE_USB_DEV_EP)

#define USB_RX_FIFO_SIZE        256
#define USB_TX0_FIFO_SIZE       64
#define USB_TX1_FIFO_SIZE       64
#define USB_TX2_FIFO_SIZE       64
#define USB_TX3_FIFO_SIZE       64
#define USB_TX4_FIFO_SIZE       64
#define USB_TX5_FIFO_SIZE       64
#define USB_TX6_FIFO_SIZE       64

#define USB_TX_FIFO(n)          *((__packed volatile uint32_t*)(USB->FIFO0D + n*1024))

#define USB_DIEPTSIZ(EPNum)     *(&USB->DIEP0TSIZ    + EPNum * 8)
#define USB_DIEPCTL(EPNum)      *(&USB->DIEP0CTL     + EPNum * 8)
#define USB_DTXFSTS(EPNum)      *(&USB->DIEP0TXFSTS  + EPNum * 8)
#define USB_DOEPTSIZ(EPNum)     *(&USB->DOEP0TSIZ    + EPNum * 8)
#define USB_DOEPCTL(EPNum)      *(&USB->DOEP0CTL     + EPNum * 8)
#define USB_DIEPINT(EPNum)      *(&USB->DIEP0INT     + EPNum * 8)
#define USB_DOEPINT(EPNum)      *(&USB->DOEP0INT     + EPNum * 8)

#define USB_EP_IN_TYPE(num)      ((USB_DIEPCTL(num) >> 18) & 3)
#define USB_EP_OUT_TYPE(num)     ((USB_DOEPCTL(num) >> 18) & 3)

// USB Driver Flags
#define USB_INITIALIZED           (1 << 0)
#define USB_POWERED               (1 << 1)
#define USB_CONNECTED             (1 << 2)

// USB Device Endpoint Flags
#define USBD_EP_FLAG_CONFIGURED   (1 << 0)
#define USBD_EP_FLAG_BUSY         (1 << 1)

// Endpoint number and index
#define EP_NUM(ep_addr)           (ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_ID(ep_addr)            (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK ? (EP_NUM(ep_addr) * 2) : ((ep_addr * 2) + 1))

// Endpoint structure
typedef struct {
  uint8_t  *buffer;
  uint32_t  bufferIndex;
  uint32_t  dataSize;
  uint32_t  maxPacketSize;
  uint8_t   flags;
} ENDPOINT;

// Static Variables
static ARM_USBD_SignalDeviceEvent_t   cbDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t cbEndpointEvent;

static uint32_t usbd_flags = 0;
static ARM_USBD_STATE usbd_state = {0, 0, 0};

         static uint8_t  setup_buf[8];
volatile static uint8_t  setup_flag = 0;

volatile static ENDPOINT ep[(USBD_EP_NUM + 1)*2];
volatile static uint32_t InEndpointNAK;

// External Variables
extern uint8_t USB_role;

// Function prototypes
static int32_t   USBD_PowerControl   (ARM_POWER_STATE state);
static uint16_t  USBD_GetFrameNumber (void);

// External Functions
extern bool USB_PinsConfigure   (uint8_t pins_mask);
extern bool USB_PinsUnconfigure (uint8_t pins_mask);
extern bool USB_PinVbusOnOff    (bool state);


// Local function

/**
  \fn          void USBD_FlushInEpFifo (uint8_t ep_addr)
  \brief       Flush IN endpoint fifo
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
*/
static void USBD_FlushInEpFifo (uint8_t ep_addr) {
  ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;

  // Flush transmit FIFO 
  USB->GRSTCTL = (USB->GRSTCTL & ~_USB_GRSTCTL_TXFNUM_MASK) |
                 (ep_addr << _USB_GRSTCTL_TXFNUM_SHIFT )    |
                  USB_GRSTCTL_TXFFLSH;
}

/**
  \fn          void USBD_Reset (void)
  \brief       Called after usbd reset interrupt to reset configuration
*/
static void USBD_Reset (void) {
  uint32_t i;

  // Clear endpoint mask registers
  USB->DOEPMSK = 0;
  USB->DIEPMSK = 0;

//  ep[EP_ID(0)].flags                                    = 0;
//  ep[EP_ID(0 | ARM_USB_ENDPOINT_DIRECTION_MASK)].flags  = 0;

  for (i = 0; i <= USBD_EP_NUM; i++) {
    if (USB_DOEPCTL(i) & USB_DOEP_CTL_EPENA) {
      USB_DOEPCTL(i)   = USB_DOEP_CTL_EPDIS | // Endpoint disable
                         USB_DOEP_CTL_SNAK;   // Endpoint set NAK
    }
    if (USB_DIEPCTL(i) & USB_DIEP_CTL_EPENA) {
      USB_DIEPCTL(i)   = USB_DIEP_CTL_EPDIS | // Endpoint disable
                         USB_DIEP_CTL_SNAK;   // Endpoint set NAK
    }
    USBD_FlushInEpFifo(i);

    // Reset endpoint resources
    ep[EP_ID(i)].buffer                                           = NULL;
    ep[EP_ID(i)].dataSize                                         = 0;
    ep[EP_ID(i)].maxPacketSize                                    = 0;
    ep[EP_ID(i)].flags                                            = 0;
    ep[EP_ID(i | ARM_USB_ENDPOINT_DIRECTION_MASK)].buffer         = NULL;
    ep[EP_ID(i | ARM_USB_ENDPOINT_DIRECTION_MASK)].dataSize       = 0;
    ep[EP_ID(i | ARM_USB_ENDPOINT_DIRECTION_MASK)].maxPacketSize  = 0;
    ep[EP_ID(i | ARM_USB_ENDPOINT_DIRECTION_MASK)].flags          = 0;
    
    // Clear IN endpoint interrupts
    USB_DIEPINT(i) = USB_DIEP_INT_XFERCOMPL   |
                     USB_DIEP_INT_EPDISBLD    |
                     USB_DIEP_INT_INTKNTXFEMP |
                     USB_DIEP_INT_INEPNAKEFF  |
                     USB_DIEP_INT_TXFEMP;

    // Clear OUT endpoint interrupts
    USB_DOEPINT(i) = USB_DOEP_INT_XFERCOMPL   |
                     USB_DOEP_INT_EPDISBLD    |
                     USB_DOEP_INT_SETUP       |
                     USB_DOEP_INT_OUTTKNEPDIS;
  }

  // Set device address to 0
  USB->DCFG       = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) | (0 << _USB_DCFG_DEVADDR_SHIFT);
  USB->DAINTMSK   =  USB_DAINTMSK_INEPMSK0 |    // Enable IN endpoint0 interrupt
                     USB_DAINTMSK_OUTEPMSK0;    // Enable OUT endpoint0 interrupt

  // Enable Setup phase done and Out transfer complete interrupt
  USB->DOEPMSK    = USB_DOEPMSK_SETUPMSK      |
                    USB_DOEPMSK_XFERCOMPLMSK;

  // Enable In transfer complete interrupt
  USB->DIEPMSK    = USB_DIEPMSK_XFERCOMPLMSK;



  // Configure FIFOs
  USB->GRXFSIZ   =  USB_RX_FIFO_SIZE   / 4;
  USB->GNPTXFSIZ = (USB_RX_FIFO_SIZE   / 4) |
                   ((USB_TX0_FIFO_SIZE / 4) << 16);

  USB->DIEPTXF1  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE) / 4) | ((USB_TX1_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF2  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE) / 4) | ((USB_TX2_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF3  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE) / 4) | ((USB_TX3_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF4  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE) / 4) | ((USB_TX4_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF5  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE) / 4) | ((USB_TX5_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF6  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE + USB_TX5_FIFO_SIZE) / 4) | ((USB_TX6_FIFO_SIZE / 4) << 16);
}

/**
\fn          void USBD_EndpointReadSet (uint8_t ep_addr)
\brief       Set Endpoint for next read.
\param[in]   ep_addr specifies Endpoint Address
              ep_addr.0..3: Address
              ep_addr.7:    Direction
*/
static void USBD_EndpointReadSet (uint8_t ep_addr) {
  uint32_t sz;
  
  // Set packet count and transfer size
  if  (ep[EP_ID(ep_addr)].dataSize > ep[EP_ID(ep_addr)].maxPacketSize) sz = ep[EP_ID(ep_addr)].maxPacketSize;
  else                                                                 sz = ep[EP_ID(ep_addr)].dataSize;
  if (ep_addr != 0) {
    USB_DOEPTSIZ(ep_addr) = (1 << _USB_DOEP_TSIZ_PKTCNT_SHIFT) | 
                             sz;
  } else {
    USB_DOEPTSIZ(ep_addr) = (1 << _USB_DOEP0TSIZ_PKTCNT_SHIFT) | 
                            (3 << _USB_DOEP0TSIZ_SUPCNT_SHIFT) |
                             sz;
  }

  // Clear NAK  and enable endpoint
  USB_DOEPCTL(ep_addr) |= USB_DOEP_CTL_EPENA | USB_DOEP_CTL_CNAK;
}

/**
  \fn          int32_t USBD_ReadFromFifo (uint8_t ep_addr, uint32_t sz)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   sz specifies data size to be read from fifo
  \return      number of data bytes read
*/
static int32_t USBD_ReadFromFifo (uint8_t ep_addr, uint32_t sz) {
  uint32_t i, residue;
  uint8_t  tmpBuf[4];

  // Check if Endpoint is activated and buffer available
  if ((USB_DOEPCTL(ep_addr) & USB_DOEP_CTL_USBACTEP) == 0) return 0;
  if (ep[EP_ID(ep_addr)].buffer == NULL)                   return 0;

  if (sz > ep[EP_ID(ep_addr)].dataSize)
    sz =ep[EP_ID(ep_addr)].dataSize;

  // Copy data from fifo
  for (i = 0; i < (uint32_t)(sz / 4); i++) {
    *((__packed uint32_t *)(ep[EP_ID(ep_addr)].buffer + ep[EP_ID(ep_addr)].bufferIndex)) = *USB->FIFO0D;
    ep[EP_ID(ep_addr)].bufferIndex += 4;
  }
  // if data size is not equal n*4
  residue = sz % 4;
  if (residue != 0) {
    *((__packed uint32_t *)tmpBuf) = *USB->FIFO0D;
    for (i = 0; i < residue; i++) {
      ep[EP_ID(ep_addr)].buffer[ep[EP_ID(ep_addr)].bufferIndex++] = tmpBuf[i];
    }
  }

  if (sz != ep[EP_ID(ep_addr)].maxPacketSize) ep[EP_ID(ep_addr)].dataSize  = 0;
  else                                        ep[EP_ID(ep_addr)].dataSize -= sz;

  return (sz);
}

/**
  \fn          void USBD_WriteToFifo (uint8_t ep_addr)
  \brief       Write data to endpoint fifo.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
*/
static void USBD_WriteToFifo (uint8_t ep_addr) {
  uint8_t  ep_num;
  uint32_t sz, i;

  ep_num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  if (ep[EP_ID(ep_addr)].dataSize > ep[EP_ID(ep_addr)].maxPacketSize)
    sz = ep[EP_ID(ep_addr)].maxPacketSize;
  else
    sz = ep[EP_ID(ep_addr)].dataSize;

  // Check if enough space in fifo
  if ((USB_DTXFSTS(ep_num) * 4) < sz) return;

  // Set transfer size and packet count
  // Set transfer size and packet count
  USB_DIEPTSIZ(ep_num) = (1 << _USB_DIEP_TSIZ_PKTCNT_SHIFT) |
                         (1 << _USB_DIEP_TSIZ_MC_SHIFT)     |
                          sz;

  // Enable endpoint and clear NAK
  USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_EPENA | USB_DIEP_CTL_CNAK;

  // Copy data to fifo
  for (i = 0; i < (uint32_t)((sz+3)/4); i++) {
    USB_TX_FIFO(ep_num) = *(__packed uint32_t *)(ep[EP_ID(ep_addr)].buffer + ep[EP_ID(ep_addr)].bufferIndex + 4*i);
  }

  ep[EP_ID(ep_addr)].bufferIndex += sz;
  ep[EP_ID(ep_addr)].dataSize    -= sz;
}


/* USB DEVICE DRIVER FUNCTIONS */

/**
  \fn          ARM_DRIVER_VERSION ARM_USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBD_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES ARM_USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_GetCapabilities (void) { return usbd_driver_capabilities; }

/**
  \fn          int32_t ARM_USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                            ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref execution_status
*/
static int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  if (usbd_flags & USB_POWERED) {
    // Device is powered - could not be re-initialized
    return ARM_DRIVER_ERROR;
  }

  if (usbd_flags & USB_INITIALIZED){
    // Driver is already initialized
    return ARM_DRIVER_OK;
  }

  // Initialize USBD Run-time Resources
  cbDeviceEvent   = cb_device_event;
  cbEndpointEvent = cb_endpoint_event;

  USB_role = ARM_USB_ROLE_DEVICE;

  // Set pin functions
  USB_PinsConfigure (ARM_USB_PIN_DP | ARM_USB_PIN_DM);

  usbd_flags = USB_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref execution_status
*/
static int32_t USBD_Uninitialize (void) {

  if (usbd_flags & USB_POWERED) {
    // Device is powered - could not be uninitialized
    return ARM_DRIVER_ERROR;
  }

  if ((usbd_flags & USB_INITIALIZED) == 0){
    // Driver not initialized
    return ARM_DRIVER_OK;
  }

  usbd_flags = 0;

  USB_role = ARM_USB_ROLE_NONE;

  // Reset pin functions
  USB_PinsUnconfigure (ARM_USB_PIN_DP | ARM_USB_PIN_DM);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBD_PowerControl (ARM_POWER_STATE state) {

  if ((usbd_flags & USB_INITIALIZED) == 0){
    // Driver is not initialized
    return ARM_DRIVER_ERROR;
  }

  switch (state) {
    case ARM_POWER_OFF:
      if ((usbd_flags & USB_POWERED) == 0)
        return ARM_DRIVER_OK;

      NVIC_DisableIRQ(USB_IRQn);                    // Disable USB interrupt

      USB->DCTL         |=  USB_DCTL_SFTDISCON;     // Soft disconnect enabled
      USB->GAHBCFG      &= ~USB_GAHBCFG_GLBLINTRMSK;// Disable global interrupt mask

      // Disable USB clock
      CMU->HFCORECLKEN0 &= ~(CMU_HFCORECLKEN0_USB | CMU_HFCORECLKEN0_USBC);

      usbd_flags = USB_INITIALIZED;
      break;

    case ARM_POWER_FULL:
      if (usbd_flags & USB_POWERED)
        return ARM_DRIVER_OK;

      CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USBC;
      CMU_ClockSelectSet( cmuClock_USBC, cmuSelect_HFCLK );
      osDelay(100);
      CMU->HFCORECLKEN0 |=  CMU_HFCORECLKEN0_USB;   // Enable USB clock

      // Wait until AHB Master state machine is in the idle condition
      while ((USB->GRSTCTL & USB_GRSTCTL_AHBIDLE) == 0);

      // Core soft reset
      USB->GRSTCTL      |=  USB_GRSTCTL_CSFTRST;
      while (USB->GRSTCTL & USB_GRSTCTL_CSFTRST);

      osDelay(10);

      USB->GAHBCFG      &= ~USB_GAHBCFG_GLBLINTRMSK;// Disable global interrupt mask
      USB->DCTL         |=  USB_DCTL_SFTDISCON;     // Soft disconnect enabled

      // Set turnaround time and force device mode
      USB->GUSBCFG   =  (USB->GUSBCFG & ~_USB_GUSBCFG_USBTRDTIM_MASK) |
                        (5 << _USB_GUSBCFG_USBTRDTIM_SHIFT)           |
                         USB_GUSBCFG_FORCEDEVMODE;
      osDelay(100);

      USB_role = ARM_USB_ROLE_DEVICE;

      // Clear interrupts
      USB->GINTSTS   =   USB_GINTSTS_CURMOD       |
                         USB_GINTSTS_SOF          |
                         USB_GINTSTS_ERLYSUSP     |
                         USB_GINTSTS_USBSUSP      |
                         USB_GINTSTS_USBRST       |
                         USB_GINTSTS_ENUMDONE     |
                         USB_GINTSTS_ISOOUTDROP   |
                         USB_GINTSTS_INCOMPISOIN  |
                         USB_GINTSTS_INCOMPLP     |
                         USB_GINTSTS_FETSUSP      |
                         USB_GINTSTS_RESETDET     |
                         USB_GINTSTS_CONIDSTSCHNG |
                         USB_GINTSTS_DISCONNINT   |
                         USB_GINTSTS_SESSREQINT   |
                         USB_GINTSTS_WKUPINT;

      USB->GINTMSK   =   USB_GINTSTS_USBSUSP      |
                         USB_GINTSTS_USBRST       |
                         USB_GINTSTS_ENUMDONE     |
                         USB_GINTSTS_RXFLVL       |
                         USB_GINTSTS_IEPINT       |
                         USB_GINTSTS_OEPINT       |
                         USB_GINTSTS_WKUPINT;

      // Enable interrupts
      NVIC_EnableIRQ(USB_IRQn);
      USB->GAHBCFG  |=   USB_GAHBCFG_GLBLINTRMSK |
                         USB_GAHBCFG_NPTXFEMPLVL;

      usbd_flags    |= USB_POWERED;

      break;
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceConnect (void) {

  if ((usbd_flags & USB_POWERED) == 0) {
    // Device not powered
    return ARM_DRIVER_ERROR;
  }

  if (usbd_flags & USB_CONNECTED) {
    // Device is already connected
    return ARM_DRIVER_OK;
  }

  USB->DCTL &= ~USB_DCTL_SFTDISCON;     // Soft disconnect disabled

  usbd_flags |= USB_CONNECTED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceDisconnect (void) {

  if ((usbd_flags & USB_POWERED) == 0) {
    // Device not powered
    return ARM_DRIVER_ERROR;
  }

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is already disconnected
    return ARM_DRIVER_OK;
  }

  USB->DCTL |= USB_DCTL_SFTDISCON;      // Soft disconnect enabled

  usbd_state.active   =  false;
  usbd_flags         &= ~USB_CONNECTED;

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBD_STATE ARM_USBD_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      Device State \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_DeviceGetState (void) {
  return usbd_state;
}

/**
  \fn          int32_t ARM_USBD_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceRemoteWakeup (void) {

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  // Remote wakeup signalling
  USB->DCTL   |=   USB_DCTL_RMTWKUPSIG;
  osDelay(5);
  USB->DCTL   &=  ~USB_DCTL_RMTWKUPSIG;
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_DeviceSetAddress (uint8_t dev_addr)
  \brief       Set USB Device Address.
  \param[in]   dev_addr  Device Address
  \return      \ref execution_status
*/
static int32_t USBD_DeviceSetAddress (uint8_t dev_addr) {

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  USB->DCFG = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) | 
              (dev_addr << _USB_DCFG_DEVADDR_SHIFT);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_ReadSetupPacket (uint8_t *setup)
  \brief       Read setup packet received over Control Endpoint.
  \param[out]  setup  Pointer to buffer for setup packet
  \return      \ref execution_status
*/
static int32_t USBD_ReadSetupPacket (uint8_t *setup) {

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  if (setup_flag == 0) {
    // No setup packet waiting
    return ARM_DRIVER_ERROR;
  }

  setup_flag = 0;
  memcpy(setup, setup_buf, 8);

  if (setup_flag) {
    // Interrupted with new setup packet
    return ARM_DRIVER_ERROR;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_EndpointConfigure (uint8_t  ep_addr,
                                                   uint8_t  ep_type,
                                                   uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type  Endpoint Type
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBD_EndpointConfigure (uint8_t ep_addr,
                                       uint8_t  ep_type,
                                       uint16_t ep_max_packet_size) {
  uint32_t val, dir;
  uint8_t num;

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  if (ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_BUSY) {
    // Endpoint is busy
    return ARM_DRIVER_ERROR_BUSY;
  }

  if (ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_CONFIGURED) {
    // Endpoint is already configured
    return ARM_DRIVER_OK;
  }

  num  = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;
  val  = ep_max_packet_size & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) dir = 16;
  else                                           dir = 0;

  // Check if endpoint is enabled 
  if ((RTE_USB_DEV_EP & (1 << num)) == 0) return ARM_DRIVER_ERROR;

  // Check if endpoint supports requested type
  switch (ep_type) {
    case ARM_USB_ENDPOINT_ISOCHRONOUS:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
    case ARM_USB_ENDPOINT_BULK:
      if ((RTE_USB_DEV_EP_BULK & (1 << (num + dir))) == 0)
        return ARM_DRIVER_ERROR;
      break;
    case ARM_USB_ENDPOINT_INTERRUPT:
      if ((RTE_USB_DEV_EP_INT  & (1 << (num + dir))) == 0)
        return ARM_DRIVER_ERROR;
      break;
    default:
      break;
  }

  // Set Endpoint resources
  ep[EP_ID(ep_addr)].buffer        = NULL;
  ep[EP_ID(ep_addr)].dataSize      = 0;
  ep[EP_ID(ep_addr)].maxPacketSize = val;

  // IN Endpoint Configuration
  if (dir) {

    // Clear IN Endpoint NAK flag
    InEndpointNAK  &= ~(1 << num);

    // Configure IN endpoint
    USB_DIEPCTL(num) = (num     <<  _USB_DIEP_CTL_TXFNUM_SHIFT) | // Fifo Number
                       (ep_type <<  _USB_DIEP_CTL_EPTYPE_SHIFT) | // Endpoint Type
                       (val);                                     // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK) {
      USB_DIEPCTL(num) |= USB_DIEP_CTL_SETD0PIDEF;
    }

    USB_DIEPCTL(num)   |= USB_DIEP_CTL_USBACTEP;  // Activate Endpoint

    if (USB_DIEPCTL(num) & USB_DIEP_CTL_EPENA) {
      USB_DIEPCTL(num) |= USB_DIEP_CTL_EPDIS;     // Disable endpoint
    }
    USB_DIEPCTL(num)   |= USB_DIEP_CTL_SNAK;      // Set Endpoint NAK

    USBD_FlushInEpFifo (num | ARM_USB_ENDPOINT_DIRECTION_MASK);

  // OUT Endpoint Configuration
  } else {

    // Configure OUT endpoint
    USB_DOEPCTL(num) = (ep_type <<  _USB_DOEP_CTL_EPTYPE_SHIFT)|  // Endpoint Type
                        USB_DOEP_CTL_SNAK                      |  // Set NAK
                        val;                                      // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK) {
      USB_DOEPCTL(num) |= USB_DOEP_CTL_SETD0PIDEF;
    }

    USB_DOEPCTL(num) |= USB_DOEP_CTL_USBACTEP;      // Activate Endpoint
  }

  // Set endpoint configured flag
  ep[EP_ID(ep_addr)].flags |= USBD_EP_FLAG_CONFIGURED;

   // Enable endpoint interrupt
  USB->DAINTMSK |= (1 << (num + dir));

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointUnconfigure (uint8_t ep_addr) {
  uint32_t num /*, IsoEpEnCnt*/;

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  if (ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_BUSY) {
    // Endpoint is busy
    return ARM_DRIVER_ERROR_BUSY;
  }

  if ((ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_CONFIGURED) == 0) {
    // Endpoint is not configured
    return ARM_DRIVER_OK;
  }

//  IsoEpEnCnt = 0;
  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  // Unconfigure IN Endpoint
  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {

    USB->DAINTMSK &= ~(1 << num);      // Disable IN EP interrupt

    if (USB_DIEPCTL(num) & USB_DIEP_CTL_EPENA)
      USB_DIEPCTL(num)  |=  USB_DIEP_CTL_EPDIS;           // Disable Endpoint
    USB_DIEPCTL(num)    |=  USB_DIEP_CTL_SNAK;            // Set Endpoint NAK

    if (num) USB_DIEPCTL(num) &= ~USB_DIEP_CTL_USBACTEP;  // Deactivate Endpoint

    InEndpointNAK &= ~(1 << num);

  // Unconfigure OUT Endpoint
  } else {
    USB->DAINTMSK &= ~(1 << (num + 16));    // Disable OUT EP interrupt

    // Set Global Out Nak
    USB->DCTL |= USB_DCTL_SGOUTNAK;
    while ((USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) == 0);

    USB_DOEPCTL(num) |= USB_DOEP_CTL_SNAK;            // Set Endpoint NAK

    if (num) {
      // Disable OUT endpoint
      if (USB_DOEPCTL(num) & USB_DOEP_CTL_EPENA) {    // If Endpoint is Enabled
        USB_DOEPCTL(num)  |=  USB_DOEP_CTL_EPDIS;     // Disable Endpoint      
        while ((USB_DOEPINT(num) & USB_DOEP_INT_EPDISBLD) == 0);
      }          
      USB_DOEPCTL(num) &= ~USB_DOEP_CTL_USBACTEP;     // Deactivate Endpoint
    }

    USB->DCTL |= USB_DCTL_CGOUTNAK;                   // Clear Global OUT NAK
  }

  // Reset IN endpoint resources
  ep[EP_ID(ep_addr)].buffer        = NULL;
  ep[EP_ID(ep_addr)].dataSize      = 0;
  ep[EP_ID(ep_addr)].maxPacketSize = 0;
  ep[EP_ID(ep_addr)].flags         = 0;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   stall  Operation
                - \b false Clear
                - \b true Set
  \return      \ref execution_status
*/
static int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall) {
  uint8_t ep_num;

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  if ((ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_CONFIGURED) == 0) {
    // Endpoint is not configured
    return ARM_DRIVER_ERROR;
  }
  if (ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_BUSY) {
    // Endpoint is not configured
    return ARM_DRIVER_ERROR_BUSY;
  }

  ep_num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  if (stall) {
    ep[EP_ID(ep_addr)].buffer      = NULL;
    ep[EP_ID(ep_addr)].dataSize    = 0;
    ep[EP_ID(ep_addr)].bufferIndex = 0;
    
    // IN Endpoint stall
    if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {

      if (USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA)
        USB_DIEPCTL(ep_num)  |= USB_DIEP_CTL_STALL | USB_DIEP_CTL_EPDIS;
      else
        USB_DIEPCTL(ep_num)  |= USB_DIEP_CTL_STALL;

      USBD_FlushInEpFifo (ep_addr);

    // OUT Endpoint Stall
    } else {

      USB->DCTL |= USB_DCTL_SGOUTNAK;               // Set Global OUT NAK
      while ((USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) == 0);

      // Stall Out endpoint
      if (USB_DOEPCTL(ep_num) & USB_DOEP_CTL_EPENA)
        USB_DOEPCTL(ep_num)  |= USB_DOEP_CTL_STALL | USB_DOEP_CTL_EPDIS;
      else
        USB_DOEPCTL(ep_num)  |= USB_DOEP_CTL_STALL;

      USB->DCTL |= USB_DCTL_CGOUTNAK;               // Clear global NAK
    }
  } else {
    // Clear IN Endpoint stall
    if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {

      if (USB_DIEPCTL(ep_num) &  USB_DIEP_CTL_EPENA) // If Endpoint enabled
        USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_EPDIS; // Disable Endpoint

      USBD_FlushInEpFifo (ep_addr);

      // Set DATA0 pid for interrupt and bulk endpoint
      if (((USB_DIEPCTL(ep_num) & _USB_DIEP_CTL_EPTYPE_MASK) >> _USB_DIEP_CTL_EPTYPE_SHIFT) > 1)
        USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_SETD0PIDEF;

      USB_DIEPCTL(ep_num) &= ~USB_DIEP_CTL_STALL;    // Clear Stall
 
    // Clear OUT Endpoint stall
    } else {
      // Set DATA0 pid for interrupt and bulk endpoint
      if (((USB_DOEPCTL(ep_num) & _USB_DOEP_CTL_EPTYPE_MASK) >> _USB_DOEP_CTL_EPTYPE_SHIFT) > 1) {
        USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SETD0PIDEF;
      }
      USB_DOEPCTL(ep_num) &= ~USB_DOEP_CTL_STALL;    // Clear stall
    }
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num)
  \brief       Read data from or Write data to USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[out]  data Pointer to buffer for data to read or with data to write
  \param[in]   num  Number of data bytes to transfer
  \return      \ref execution_status
*/
int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num) {
  uint8_t ep_num;

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  if ((ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_CONFIGURED) == 0) {
    // Endpoint is not configured
    return ARM_DRIVER_ERROR;
  }

  if (ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_BUSY) {
    // Endpoint is busy
    return ARM_DRIVER_ERROR_BUSY;
  }

  // Set endpoint busy
  ep[EP_ID(ep_addr)].flags |= USBD_EP_FLAG_BUSY;

  ep[EP_ID(ep_addr)].bufferIndex = 0;
  ep[EP_ID(ep_addr)].buffer      = data;
  ep[EP_ID(ep_addr)].dataSize    = num;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep_num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

#if (USB_INT_IN_EP != 0)
    if (USB_EP_IN_TYPE(ep_num) == ARM_USB_ENDPOINT_INTERRUPT) {
      InEndpointNAK    |=  1 << ep_num;
      USB_DIEPCTL(num) |=  USB_DIEP_CTL_SNAK;           // Set NAK
      USB->DIEPMSK     |=  USB_DIEPMSK_INEPNAKEFFMSK;   // Enable NAK effective interrupt
    } else
#endif
    {
      USBD_WriteToFifo(ep_addr);
    }

  } else {
    USBD_EndpointReadSet(ep_addr);
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t ARM_USBD_EndpointTransferGetResult (uint8_t ep_addr)
  \brief       Get result of USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      number of successfully transfered data bytes
*/
uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr) {
  return ep[EP_ID(ep_addr)].bufferIndex;
}

/**
  \fn          int32_t ARM_USBD_EndpointTransferAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransferAbort (uint8_t ep_addr) {
  uint8_t ep_num;

  if ((usbd_flags & USB_CONNECTED) == 0) {
    // Device is not connected
    return ARM_DRIVER_ERROR;
  }

  if ((ep[EP_ID(ep_addr)].flags & USBD_EP_FLAG_CONFIGURED) == 0) {
    // Endpoint is not configured
    return ARM_DRIVER_ERROR;
  }

  ep_num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  ep[EP_ID(ep_addr)].buffer      = NULL;
  ep[EP_ID(ep_addr)].dataSize    = 0;
  ep[EP_ID(ep_addr)].bufferIndex = 0;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {

    if (USB_DIEPCTL(ep_num) &  USB_DIEP_CTL_EPENA)    // If endpoint enabled
      USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_EPDIS;    // disable endpoint
    USB_DIEPCTL(ep_num)     |= USB_DIEP_CTL_SNAK;     // Set NAK

    USBD_FlushInEpFifo (ep_addr);

  } else {
    if (USB_DOEPCTL(ep_addr) &  USB_DOEP_CTL_EPENA)   // If endpoint enabled
      USB_DOEPCTL(ep_addr)   |= USB_DOEP_CTL_EPDIS;   // Disable endpoint
    USB_DOEPCTL(ep_addr)     |= USB_DOEP_CTL_SNAK;    // Set NAK
  }

  // Clear endpoint busy
  ep[EP_ID(ep_addr)].flags &= ~USBD_EP_FLAG_BUSY;

  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t ARM_USBD_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_GetFrameNumber (void) {
  return ((USB->DSTS & _USB_DSTS_SOFFN_MASK) >> _USB_DSTS_SOFFN_SHIFT);
}

/**
  \fn          void USBD_IRQ (uint32_t gintsts)
  \brief       USB Device Interrupt Routine (IRQ).
*/
void USBD_IRQ (uint32_t gintsts) {
  uint32_t ep_int, num, val, msk, sz, dev_evt;

  // Reset device event variable
  dev_evt = 0;

// Reset interrupt
  if (gintsts & USB_GINTSTS_USBRST) {
    USBD_Reset();
    dev_evt |= ARM_USBD_EVENT_RESET;
    cbDeviceEvent(dev_evt);
    USB->GINTSTS = USB_GINTSTS_USBRST;
  }

// Suspend interrupt
  if (gintsts & USB_GINTSTS_USBSUSP) {
    usbd_state.active = false;
    dev_evt |= ARM_USBD_EVENT_SUSPEND;
    cbDeviceEvent(dev_evt);
    USB->PCGCCTL    |= USB_PCGCCTL_STOPPCLK;
    USB->GINTSTS = USB_GINTSTS_USBSUSP;
  }

// Resume interrupt
  if (gintsts & USB_GINTSTS_WKUPINT) {
    usbd_state.active = true;
    dev_evt |= ARM_USBD_EVENT_RESUME;
    cbDeviceEvent(dev_evt);
    USB->PCGCCTL    &= ~USB_PCGCCTL_STOPPCLK;
    USB->GINTSTS = USB_GINTSTS_WKUPINT;
  }

// Speed enumeration completed
  if (gintsts & USB_GINTSTS_ENUMDONE) {
    usbd_state.speed     = ARM_USB_SPEED_FULL;
    usbd_state.active    = true;

    USB->DCTL    |= USB_DCTL_CGNPINNAK;      // Clear global IN NAK
    USB->DCTL    |= USB_DCTL_CGOUTNAK ;      // clear global OUT NAK
    USB->GINTSTS  = USB_GINTSTS_ENUMDONE;
  }

  if (dev_evt && cbDeviceEvent) cbDeviceEvent(dev_evt);

  if (gintsts & USB_GINTSTS_RXFLVL) {
    USB->GINTMSK &= ~USB_GINTSTS_RXFLVL;

    val =  USB->GRXSTSP;
    num =  val & 0x0F;
    sz  = (val >> 4) & 0x7FF;

    switch ((val >> 17) & 0x0F) {
      // Setup packet
      case 6:
        // Read setup packet
        *(__packed uint32_t *)(setup_buf)     = *USB->FIFO0D;
        *(__packed uint32_t *)(setup_buf + 4) = *USB->FIFO0D;

        // Analyze Setup packet for SetAddress
        if (setup_buf[0] == 0) {
          if (setup_buf[1] == 5)
            USBD_DeviceSetAddress(setup_buf[2]);
        }

        setup_flag = 1;
      break;

      // OUT packet
      case 2:
        USBD_ReadFromFifo(num, sz);
        break;

      // Global OUT NAK
      case 1:
        break;

      // OUT transfer completed
      case 3:
        break;
      
      // SETUP transaction completed
      case 4:
        break;

      default:
        break;
    }
    USB->GINTMSK |= USB_GINTSTS_RXFLVL;
  }

// OUT Packet
  if (gintsts & USB_GINTSTS_OEPINT) {
    msk = (((USB->DAINT & USB->DAINTMSK) >> 16) & 0xFFFF);
    num = 0;

    do {
      if ((msk >> num) & 1) {
        ep_int = USB_DOEPINT(num) & USB->DOEPMSK;

        // Setup phase done interrupt
        if (ep_int & USB_DOEP_INT_SETUP) {
          ep[EP_ID(num)].dataSize = 0;
          USB_DOEPINT(num) = USB_DOEP_INT_SETUP;
          cbEndpointEvent(num, ARM_USBD_EVENT_SETUP);
        }

        // Transfer complete interrupt
        if (ep_int & USB_DOEP_INT_XFERCOMPL) {
          USB_DOEPINT(num) = USB_DOEP_INT_XFERCOMPL;
          if (USB_EP_OUT_TYPE(num) != ARM_USB_ENDPOINT_ISOCHRONOUS) {
            if (ep[EP_ID(num)].dataSize != 0)
              USBD_EndpointReadSet(num);
            else {
              ep[EP_ID(num)].flags  &= ~USBD_EP_FLAG_BUSY;
              cbEndpointEvent(num, ARM_USBD_EVENT_OUT);
            }
          }
        }
      }
      num++;
    } while (msk >> num);
  }

// IN Packet
  if (gintsts & USB_GINTSTS_IEPINT) {
    msk = (USB->DAINT & USB->DAINTMSK & 0xFFFF);
    num = 0;

    do {
      if ((msk >> num) & 1) {
        ep_int = USB_DIEPINT(num) & USB->DIEPMSK;

    // IN endpoint NAK effective
#if (USB_INT_IN_EP)
       if (ep_int & USB_DIEP_INT_INEPNAKEFF) {
          if (USB_EP_IN_TYPE(num) == ARM_USB_ENDPOINT_INTERRUPT) {
            if (InEndpointNAK & (1 << num)) {
              InEndpointNAK &= ~ (1 << num);
              // if no more forced NAKs, disable IN NAK effective interrupt
              if (InEndpointNAK == 0) USB->DIEPMSK &= ~USB_DIEPMSK_INEPNAKEFFMSK;

              // If Data available, write Data
              if ((ep[EP_ID(num | ARM_USB_ENDPOINT_DIRECTION_MASK)].buffer) &&
                  ((ep[EP_ID(num | ARM_USB_ENDPOINT_DIRECTION_MASK)].bufferIndex == 0))) {
                USBD_WriteToFifo(num | ARM_USB_ENDPOINT_DIRECTION_MASK);
              }
            }
          }
          USB_DIEPINT(num) = USB_DIEP_INT_INEPNAKEFF;
        }
#endif

        // Transmit completed
        if (ep_int & USB_DIEP_INT_XFERCOMPL) {
          USB_DIEPINT(num) = USB_DIEP_INT_XFERCOMPL;
          if (ep[EP_ID(num | ARM_USB_ENDPOINT_DIRECTION_MASK)].dataSize == 0) {
            ep[EP_ID(num | ARM_USB_ENDPOINT_DIRECTION_MASK)].buffer =  NULL;
            ep[EP_ID(num | ARM_USB_ENDPOINT_DIRECTION_MASK)].flags  &= ~USBD_EP_FLAG_BUSY;
            cbEndpointEvent(num | ARM_USB_ENDPOINT_DIRECTION_MASK, ARM_USBD_EVENT_IN);
          } else {
            USBD_WriteToFifo(num | ARM_USB_ENDPOINT_DIRECTION_MASK);
          }
        }
      }
      num++;
    } while (msk >> num);
  }
}

ARM_DRIVER_USBD Driver_USBD0 = {
  USBD_GetVersion,
  USBD_GetCapabilities,
  USBD_Initialize,
  USBD_Uninitialize,
  USBD_PowerControl,
  USBD_DeviceConnect,
  USBD_DeviceDisconnect,
  USBD_DeviceGetState,
  USBD_DeviceRemoteWakeup,
  USBD_DeviceSetAddress,
  USBD_ReadSetupPacket,
  USBD_EndpointConfigure,
  USBD_EndpointUnconfigure,
  USBD_EndpointStall,
  USBD_EndpointTransfer,
  USBD_EndpointTransferGetResult,
  USBD_EndpointTransferAbort,
  USBD_GetFrameNumber
};
